/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * a sample program to use Splus library
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#include "Splus.h"


/* the main function */
int main(int argc, char *argv[]) {
  double t1 = 0.0, t2 = 0.0, t3 = 0.0, t4 = 0.0;
  SP_SparseMatrix *A = NULL;
  SP_DenseMatrix *B = NULL;
  int i, me_no, nproc;
  int *perm_c;		/* column permutation vector from mmd ordering */
  int *perm_r;		/* row permutation vector to retain zero-free diagonal */

  MPI_Init (&argc, &argv);

  if (argc!=1) {
    fprintf(stderr, "usage: no parameter\n");
    exit(0);
  }

  MPI_Comm_size (MPI_COMM_WORLD, &nproc) ;
  MPI_Comm_rank (MPI_COMM_WORLD, &me_no) ;

  if (me_no == 0) {
    t1 = MPI_Wtime ();

    /* Initialize matrix A */
    A = (SP_SparseMatrix*) malloc (sizeof(SP_SparseMatrix));
    A->order = 5;
    A->numNZ = 12;
    A->coe = (ELEMTYPE*) malloc (A->numNZ*sizeof(ELEMTYPE));
    A->coe[0] = 19.0; A->coe[1] = 12.0; A->coe[2] = 12.0; A->coe[3] = 21.0; 
    A->coe[4] = 12.0; A->coe[5] = 12.0; A->coe[6] = 21.0; A->coe[7] = 16.0; 
    A->coe[8] = 21.0; A->coe[9] =  5.0; A->coe[10]= 21.0; A->coe[11]= 18.0;
    A->rowIndex = (int*) malloc (A->numNZ*sizeof(int));
    A->rowIndex[0] = 0; A->rowIndex[1] = 1; A->rowIndex[2] = 4; A->rowIndex[3] = 1;
    A->rowIndex[4] = 2; A->rowIndex[5] = 4; A->rowIndex[6] = 0; A->rowIndex[7] = 2;
    A->rowIndex[8] = 0; A->rowIndex[9] = 3; A->rowIndex[10]= 3; A->rowIndex[11]= 4;
    A->colStart = (int*) malloc ((A->order+1)*sizeof(int));
    A->colStart[0] = 0; A->colStart[1] = 3; A->colStart[2] = 6; 
    A->colStart[3] = 8; A->colStart[4] = 10; A->colStart[5] = 12;

    /* Initialize right-hand side B */
    B = (SP_DenseMatrix*) malloc (sizeof(SP_DenseMatrix));
    if (B==NULL) {
      fprintf(stderr, "out of memory\n");
      exit(0);
    }
    B->rowNum = A->order; B->colNum = 1;
    B->element = (ELEMTYPE*) malloc (B->rowNum*B->colNum*sizeof(ELEMTYPE));
    if (B->element==NULL) {
      fprintf(stderr, "out of memory\n");
      exit(0);
    }
    for (i=0; i<B->rowNum; i++) 
      B->element[i] = 1.0;

    t2 = MPI_Wtime();

    /* Ordering matrix using column MMD, called only by MPI node 0 */
    SP_Ordering (1, A, &perm_c, &perm_r);

    t3 = MPI_Wtime();
  }

  /* Solve sparse linear system A*X=B, called by all MPI nodes.
   * The solution overwrites B at completion.
   */
  SP_Solve(A, B, perm_c, perm_r, me_no, nproc); 

  if (me_no == 0) {
    t4 = MPI_Wtime();

    /* free the intermediate space */
    SP_DestroySparseMatrix(A);
    free(perm_c);
    free(perm_r);

    /* print the result */
    fprintf (stdout, "## Result:  "); 
    for (i=0; i<B->rowNum; i++) fprintf (stdout, " %8.3f", B->element[i]); 
    fprintf (stdout, "\n");

    /* free the space for B and X */
    if (B->element!=NULL) free(B->element);
    if (B!=NULL) free(B);

    /* print performance statistics */
    fprintf(stdout, "## Performance stats: reading -%8.3fsec, ordering -%8.3fsec, solving -%8.3fsec\n", t2-t1, t3-t2, t4-t3);
  }

  MPI_Finalize();
 
  return 1;
}
