/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * a sample program to use Splus library, it takes input matrix file in the ij-value format
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#include "Splus.h"


/* verify the result of A*X=B */
double VerifyResult(SP_SparseMatrix *A, SP_DenseMatrix *X, SP_DenseMatrix *B);


/* the main function */
int main(int argc, char *argv[]) {
  double t1 = 0.0, t2 = 0.0, t3 = 0.0, t4 = 0.0;
  char *inputmatrix;
  SP_SparseMatrix *A = NULL;
  SP_DenseMatrix *B = NULL;
  SP_DenseMatrix *X = NULL;
  int i, me_no, nproc;
  int *perm_c;		/* column permutation vector from mmd ordering */
  int *perm_r;		/* row permutation vector to retain zero-free diagonal */

  MPI_Init (&argc, &argv);

  if (argc!=2) {
    fprintf(stderr, "usage: %s <matrix_file>\n", argv[0]);
    exit(0);
  }

  inputmatrix = argv[1];

  MPI_Comm_size (MPI_COMM_WORLD, &nproc) ;
  MPI_Comm_rank (MPI_COMM_WORLD, &me_no) ;

  if (me_no == 0) {
    t1 = MPI_Wtime ();

    /* Initialize matrix A */
    A = SP_ReadMatrixFromHB(inputmatrix);

    /* Initialize right-hand side B */
    B = (SP_DenseMatrix*) malloc (sizeof(SP_DenseMatrix));
    if (B==NULL) {
      fprintf(stderr, "out of memory\n");
      exit(0);
    }
    B->rowNum = A->order; B->colNum = 1;
    B->element = (ELEMTYPE*) malloc (B->rowNum*B->colNum*sizeof(ELEMTYPE));
    if (B->element==NULL) {
      fprintf(stderr, "out of memory\n");
      exit(0);
    }
    for (i=0; i<B->rowNum; i++) 
      B->element[i] = 1.0;

    /* Initialize the solution X (the same as B) */
    X = (SP_DenseMatrix*) malloc (sizeof(SP_DenseMatrix));
    if (X==NULL) {
      fprintf(stderr, "out of memory\n");
      exit(0);
    }
    X->rowNum = B->rowNum; X->colNum = 1;
    X->element = (ELEMTYPE*) malloc (X->rowNum*X->colNum*sizeof(ELEMTYPE));
    if (X->element==NULL) {
      fprintf(stderr, "out of memory\n");
      exit(0);
    }
    for (i=0; i<X->rowNum; i++) 
      X->element[i] = 1.0;

    t2 = MPI_Wtime();

    /* Ordering matrix using column MMD, called only by MPI node 0 */
    SP_Ordering (1, A, &perm_c, &perm_r);

    t3 = MPI_Wtime();
  }

  /* Solve sparse linear system A*X=B, called by all MPI nodes.
   * X is both input and output.  As input, it is initialized as the right-hand side B.
   */
  SP_Solve(A, X, perm_c, perm_r, me_no, nproc); 

  if (me_no == 0) {
    t4 = MPI_Wtime();

    /* free the intermediate space */
    SP_DestroySparseMatrix(A);
    free(perm_c);
    free(perm_r);

    /* verify the result */
    A = SP_ReadMatrixFromHB(inputmatrix);
    fprintf(stdout, "## Error: %g\n", VerifyResult(A, X, B));
    SP_DestroySparseMatrix(A);

    /* free the space for B and X */
    if (B->element!=NULL) free(B->element);
    if (B!=NULL) free(B);
    if (X->element!=NULL) free(X->element);
    if (X!=NULL) free(X);

    /* print performance statistics */
    fprintf(stdout, "## Performance stats: reading -%8.3fsec, ordering -%8.3fsec, solving -%8.3fsec\n", t2-t1, t3-t2, t4-t3);
  }

  MPI_Finalize();
 
  return 1;
}


/* We first calculate B2=A*X, then we return |B2-B|/|B|. */
double VerifyResult(SP_SparseMatrix *A, SP_DenseMatrix *X, SP_DenseMatrix *B) {
  SP_DenseMatrix *B2;
  SP_DenseMatrix *B2abs;
  int curCol, i;
  ELEMTYPE maxerror = 0;

  /* initialize the B2 matrix */
  B2 = (SP_DenseMatrix*) malloc (sizeof(SP_DenseMatrix));
  if (B2==NULL) {
    fprintf(stderr, "out of memory\n");
    exit(0);
  }
  B2->rowNum = X->rowNum; B2->colNum = 1;
  B2->element = (ELEMTYPE*) malloc(B2->rowNum*B2->colNum*sizeof(ELEMTYPE));
  if (B2->element==NULL) {
    fprintf(stderr, "out of memory\n");
    exit(0);
  }
  for (i=0; i<B2->rowNum; i++)
    B2->element[i] = 0.0;

  /* initialize the B2abs matrix */
  B2abs = (SP_DenseMatrix*) malloc (sizeof(SP_DenseMatrix));
  if (B2abs==NULL) {
    fprintf(stderr, "out of memory\n");
    exit(0);
  }
  B2abs->rowNum = X->rowNum; B2abs->colNum = 1;
  B2abs->element = (ELEMTYPE*) malloc(B2abs->rowNum*B2abs->colNum*sizeof(ELEMTYPE));
  if (B2abs->element==NULL) {
    fprintf(stderr, "out of memory\n");
    exit(0);
  }
  for (i=0; i<B2abs->rowNum; i++)
    B2abs->element[i] = 0.0;

  curCol = 1;
  for (i=0; i<A->numNZ; i++) {
    int col = -1;
    int row = -1;
    ELEMTYPE ele = -1;

    if (i == A->colStart[curCol]) curCol++;
    col = curCol-1;
    row = A->rowIndex[i];
    ele = A->coe[i];

    B2->element[row] += ele * X->element[col];
    B2abs->element[row] += fabs(ele * X->element[col]);
  }
	
  for (i=0; i<B2->rowNum; i++) {
    double thiserror = fabs(B2->element[i]-B->element[i])/(fabs(B2abs->element[i])+fabs(B->element[i]));
    if (thiserror>maxerror)
      maxerror = thiserror;
  }
 
  if (B2->element!=NULL) free(B2->element);
  if (B2!=NULL) free(B2);

  if (B2abs->element!=NULL) free(B2abs->element);
  if (B2abs!=NULL) free(B2abs);

  return maxerror;
}
