/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * Comm.c - Communication procedures in LU factorization
 *
 */


#include <stdio.h>
#include <string.h>

#include "struct.h"
#include "functions.h"
#include "globals.h"

extern ELEMTYPE *SP_Rbuffer, *SP_Cbuffer;
extern INDEXTYPE *SP_Rindexbuf, *SP_Cindexbuf;
extern PATTYPE *SP_Rpatternbuf, *SP_Cpatternbuf;
extern rowID *SP_PivotVector;		/* Pivot Vector */
extern int SP_cindbufsize, SP_rindbufsize;
extern INDEXTYPE SP_maxNumColSendBlks, SP_maxNumRowSendBlks;

static ELEMTYPE *_tmpLinv;
static PATTYPE *_CpatternbufOut, *_RpatternbufOut;

struct pendingreq_t {
  int reqcnt;
  MPI_Request *reqs;
  void *spacetofree;
  size_t sizetofree;
  struct pendingreq_t *next;
};

static struct pendingreq_t *_preqlist;

/* get data structure for a new pending request */
static struct pendingreq_t *_onependingreq(int reqcnt) {
  struct pendingreq_t *_tmp = _preqlist;
  _preqlist = (struct pendingreq_t*) SP_Malloc(sizeof(struct pendingreq_t));
  _preqlist->reqcnt = reqcnt;
  if (reqcnt>0)
    _preqlist->reqs = (MPI_Request*) SP_Malloc(reqcnt*sizeof(MPI_Request));
  else
    _preqlist->reqs = NULL;
  _preqlist->next = _tmp;
  return _preqlist;
}

/* garbage collection */
static void _gcpendingreq() {
  int flag;
  MPI_Status status;
  struct pendingreq_t **_ptmp = &_preqlist;
  while (*_ptmp!=NULL) {
    int complete = 1;
    int i;
    for (i=0; i<(*_ptmp)->reqcnt; i++) 
      if ((*_ptmp)->reqs[i]!=MPI_REQUEST_NULL) {
	MPI_Test(&(*_ptmp)->reqs[i], &flag, &status);
	if (flag) {
	  ASSERT((*_ptmp)->reqs[i]==MPI_REQUEST_NULL);
	}
	else {
	  complete = 0;
	  break;
	}
      }
    if (complete) {
      struct pendingreq_t *_tmp = *_ptmp;
      SP_Free(_tmp->reqs);
      if (_tmp->spacetofree!=NULL)
        SP_FreeSize(_tmp->spacetofree, _tmp->sizetofree);
      *_ptmp = _tmp->next;
      SP_Free(_tmp);
    }
    else
      *_ptmp = (*_ptmp)->next;
  }
}

void SP_InitComm() {
  _tmpLinv = (ELEMTYPE*) SP_Malloc (SP_BlkSize*SP_BlkSize*sizeof(ELEMTYPE));
  _CpatternbufOut = (PATTYPE *) SP_Malloc (SP_cindbufsize*sizeof(PATTYPE));
  _RpatternbufOut = (PATTYPE *) SP_Malloc (SP_rindbufsize*sizeof(PATTYPE));

  _preqlist = NULL;
}

void SP_FinalComm() {
  MPI_Status status;
  struct pendingreq_t *_tmp, **_ptmp;

  SP_Free (_tmpLinv);
  SP_Free (_CpatternbufOut);
  SP_Free (_RpatternbufOut);

  _ptmp = &_preqlist;
  while (*_ptmp!=NULL) {
    int i;
    for (i=0; i<(*_ptmp)->reqcnt; i++)
      if ((*_ptmp)->reqs[i]!=MPI_REQUEST_NULL) {
        MPI_Wait(&(*_ptmp)->reqs[i], &status);
      }
    _tmp = *_ptmp;
    SP_Free(_tmp->reqs);
    if (_tmp->spacetofree!=NULL)
      SP_FreeSize(_tmp->spacetofree, _tmp->sizetofree);
    *_ptmp = _tmp->next;
    SP_Free(_tmp);
  }
  ASSERT(_preqlist==NULL);
}

/* broadcast to all others block[k+1,...,snOrder][k] */
void SP_send_col_MPI(SP_SNmatrix_struct *SNma, INDEXTYPE k) {
  int pe, dsize = DSIZE(SNma, k);
  INDEXTYPE tc, tempendc;  

  _gcpendingreq();

  tc = SNma->LstartCol[CINDEX(k)];
  if (SNma->LindexCol[tc] == k) ++tc;
  tempendc = SNma->LstartCol[CINDEX(k)+1];

  if (tempendc > tc) {
    register int i, bsize, numblocks=0;
    INDEXTYPE stc = -1, endc = -1;
    register ELEMTYPE *tmpp;
    PATTYPE *patstart = NULL, *pat = _CpatternbufOut;
    for (; tc < tempendc; ++tc)
      if ((tmpp=SNma->LnodeCol[tc]) != NULL) {
	for (i=0, bsize=SNma->LsubRowNumInBlk[tc]*dsize; i < bsize; ++i)
	  if (tmpp[i] < -SP_Epsilon || tmpp[i] > SP_Epsilon)
	    break;

	if (i==bsize) {
 	  SP_FreeSize(tmpp, bsize*sizeof(ELEMTYPE));
 	  SNma->LnodeCol[tc] = NULL;
	  *pat++ = 0;
	}
        else {
	  if (!numblocks++) {
	    patstart = pat;
	    stc = tc;
	  }
	  *pat++ = SNma->LsubRow[tc];
	  endc = tc;
	}
      }
      else
	*pat++ = 0;

    ++endc; /* we want endc to be one after the last block */

    if (numblocks>0) {
      struct pendingreq_t *_preq;
      int _preqcnt;

      /* use two rounds (patterns sent first) for better caching and less wait time on the other PEs */
      i = endc-stc;
      for (pe = SP_me_rno; pe<SP_nproc; pe+=SP_rp)
	if (SP_me_no != pe) {
	  MPI_Send(patstart, i, PATTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD);
	  MPI_Send(SNma->LindexCol+stc, i, INDEXTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD);
	}

      /* pack and send the columns */
      bsize = 0;
      for (tc = stc; tc < endc; ++tc)
	if (SNma->LnodeCol[tc]!=NULL)
	  bsize += SNma->LsubRowNumInBlk[tc]*dsize;
      tmpp = (ELEMTYPE *) SP_Malloc(bsize*sizeof(ELEMTYPE));
      bsize = 0;
      for (tc = stc; tc < endc; ++tc)
	if (SNma->LnodeCol[tc]!=NULL) {
	  memcpy(&tmpp[bsize], SNma->LnodeCol[tc], SNma->LsubRowNumInBlk[tc]*dsize*sizeof(ELEMTYPE));
	  bsize += SNma->LsubRowNumInBlk[tc]*dsize;
	}
      _preq = _onependingreq(SP_cp-1);
      _preq->spacetofree = tmpp;
      _preq->sizetofree = bsize*sizeof(ELEMTYPE);
      _preqcnt = 0;
      for (pe = SP_me_rno; pe<SP_nproc; pe+=SP_rp)
	if (SP_me_no != pe)
	  /* use Isend to avoid deadlock */
	  MPI_Isend(tmpp, bsize, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &_preq->reqs[_preqcnt++]);
      /* since we do nonblocking send, we free the space only when the send is completed */
      /* SP_FreeSize(tmpp, bsize*sizeof(ELEMTYPE)); */
      return;
    }
  }

  /* if we started out with no blocks, just send out nothing */
  for (pe = SP_me_rno; pe<SP_nproc; pe+=SP_rp)
    if (SP_me_no != pe)
      MPI_Send(SNma->LsubRow, 0, PATTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD);
}

/* get block[k+1,...,snOrder][k] */
INDEXTYPE SP_get_col_MPI(SP_SNmatrix_struct *SNma, INDEXTYPE k) {
  MPI_Status SP_status;
  int i, num, size, pe=COWNER(k)*SP_rp+SP_me_rno;
  ELEMTYPE *p = SP_Cbuffer;

  MPI_Recv(SP_Cpatternbuf, SP_cindbufsize, PATTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
  MPI_Get_count(&SP_status, INDEXTYPE_MPI, &num);

  if (num>0) {
    MPI_Recv(SP_Cindexbuf, SP_cindbufsize, INDEXTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
    size = 0;
    for(i=0; i < num; ++i)
      if (SP_Cpatternbuf[i]) 
	size += SP_OrderOfInt(SP_Cpatternbuf[i])*DSIZE(SNma, k);
    MPI_Recv(p, size, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
  }

  return num;
}

/* broadcast to all others block[k][k+1,...,snOrder] */
void SP_send_row_MPI(SP_SNmatrix_struct *SNma, INDEXTYPE k) {
  int pe, dsize = DSIZE(SNma, k);
  INDEXTYPE tr, tempendr;

  tr = SNma->UstartRow[RINDEX(k)];
  tempendr = SNma->UstartRow[RINDEX(k)+1];

  if (tempendr > tr) {
    INDEXTYPE str = -1, endr = -1;
    register int i, bsize, numblocks=0;
    register ELEMTYPE *tmpp;
    PATTYPE *patstart = NULL, *pat = _RpatternbufOut;
    for (; tr < tempendr; ++tr)
      if ((tmpp=SNma->UnodeRow[tr]) != NULL) {
	for (i=0, bsize=SNma->UsubColNumInBlk[tr]*dsize; i < bsize; ++i)
	  if (tmpp[i] < -SP_Epsilon || tmpp[i] > SP_Epsilon)
	    break;

	if (i==bsize) {
 	  SP_FreeSize(tmpp, bsize*sizeof(ELEMTYPE));
 	  SNma->UnodeRow[tr] = NULL;
	  *pat++ = 0;
	}
        else {
	  if (!numblocks++) {
	    patstart = pat;
	    str = tr;
	  }
	  *pat++ = SNma->UsubCol[tr];
	  endr = tr;
	}
      }
      else
	*pat++ = 0;

    ++endr; /* we want endc to be one after the last block */

    if (numblocks>0) {
      /* use two rounds (patterns sent first) for better caching and less wait time on the other PEs */
      i = endr-str;
      for (pe = SP_me_cno*SP_rp; pe<(SP_me_cno+1)*SP_rp; pe++)
	if (SP_me_no != pe) {
	  MPI_Send(patstart, i, PATTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD);
	  MPI_Send(SNma->UindexRow+str, i, INDEXTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD);
	}

      /* pack and send the rows */
      bsize = 0;
      for (tr = str; tr < endr; ++tr)
	if (patstart[tr-str])
	  bsize += SNma->UsubColNumInBlk[tr]*dsize;
      tmpp = (ELEMTYPE *) SP_Malloc(bsize*sizeof(ELEMTYPE));
      bsize = 0;
      for (tr = str; tr < endr; ++tr)
	if (patstart[tr-str]) {
	  memcpy(&tmpp[bsize], SNma->UnodeRow[tr], SNma->UsubColNumInBlk[tr]*dsize*sizeof(ELEMTYPE));
	  bsize += SNma->UsubColNumInBlk[tr]*dsize;
	}
      for (pe = SP_me_cno*SP_rp; pe<(SP_me_cno+1)*SP_rp; pe++)
	if (SP_me_no != pe)
	  MPI_Send(tmpp, bsize, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD);
      SP_FreeSize(tmpp, bsize*sizeof(ELEMTYPE));
      return;
    }
  }

  /* if we started out with no blocks, just send out nothing */
  for (pe = SP_me_cno*SP_rp; pe<(SP_me_cno+1)*SP_rp; pe++)
    if (SP_me_no != pe)
      MPI_Send(SNma->UsubCol, 0, PATTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD);
}

/* get block[k][k+1,...,snOrder] */
INDEXTYPE SP_get_row_MPI(SP_SNmatrix_struct *SNma, INDEXTYPE k) {
  MPI_Status SP_status;
  int i, num, size, pe = SP_me_cno*SP_rp+ROWNER(k);
  ELEMTYPE *p = SP_Rbuffer;

  MPI_Recv(SP_Rpatternbuf, SP_rindbufsize, PATTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
  MPI_Get_count(&SP_status, PATTYPE_MPI, &num);

  if (num>0) {
    MPI_Recv(SP_Rindexbuf, SP_rindbufsize, INDEXTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
    size = 0;
    for(i=0; i<num; ++i)
      if (SP_Rpatternbuf[i])
	size += SP_OrderOfInt(SP_Rpatternbuf[i])*DSIZE(SNma, k);
    MPI_Recv(p, size, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
  }

  return num;
}

/* broadcast to all others the pivot for k */
void SP_send_pivot_MPI(SP_SNmatrix_struct *SNma, INDEXTYPE k) {
  int pe;

  if (SP_cp==1) return;

  for (pe = SP_me_rno; pe<SP_nproc; pe+=SP_rp)
    if (SP_me_no != pe)
      MPI_Send(SP_PivotVector+SNma->Sstart[k], DSIZE(SNma, k)*sizeof(rowID), MPI_CHAR, pe, SP_MPI_TAG, MPI_COMM_WORLD);
}

/* get pivot from COWNER(k), start col Irecvs */
void SP_get_pivot_MPI(SP_SNmatrix_struct *SNma, INDEXTYPE k) {
  MPI_Status SP_status;

  MPI_Recv(SP_PivotVector+SNma->Sstart[k], DSIZE(SNma, k)*sizeof(rowID), MPI_CHAR, COWNER(k)*SP_rp+SP_me_rno, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
}

/* send L inverse to all members of row ROWNER(k) */
void SP_send_linv_MPI(SP_SNmatrix_struct *SNma, INDEXTYPE k) {
  int pe, size=DSIZE(SNma, k), info;

  memcpy (_tmpLinv, SNma->LnodeCol[SNma->LstartCol[CINDEX(k)]], sizeof(ELEMTYPE)*size*size);
  spTRTRI (SP_ftcL, SP_ftcU, &size, _tmpLinv, &size, &info);

  for(pe=ROWNER(k); pe<SP_nproc; pe+=SP_rp)
    MPI_Send(_tmpLinv, size*size, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD);
}
