/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * Ordering.c - ordering routine
 *
 */


#include <strings.h>

#include "struct.h"
#include "functions.h"
#include "globals.h"


int genmmd_ (int*, int*, int*, int*, int*, int*, int*, int*, int*, int*, int*, int*);

static void
_construct_at_times_a (
       const int m,      /* INPUT - number of rows in matrix A. */
       const int n,      /* INPUT - number of columns in matrix A. */
       const int nz,     /* INPUT - number of nonzeros in matrix A */
       int *colptr,      /* INPUT - column pointer of size n+1 for matrix A. */
       int *rowind,      /* INPUT - row indices of size nz for matrix A. */
       int *atanz,       /* OUTPUT - on exit, returns the actual number of
                            nonzeros in matrix A'*A. */
       int **ata_colptr, /* OUTPUT - size n+1 */
       int **ata_rowind  /* OUTPUT - size *atanz */
       )
/*
 * Purpose
 * =======
 *
 * Form the structure of A'*A. A is an m-by-n matrix in column oriented
 * format represented by (colptr, rowind). The output A'*A is in column
 * oriented format (symmetrically, also row oriented), represented by
 * (ata_colptr, ata_rowind).
 *
 * This routine is modified from getata routine in SuperLU, which in turn 
 * is modified from GETATA routine by Tim Davis.
 * The complexity of this algorithm is: SUM_{i=1,m} r(i)^2,
 * i.e., the sum of the square of the row counts.
 */
{
  register int i, j, k, col, num_nz, ti, trow;
  int *marker, *thiscol, *b_colptr, *b_rowind;
  int *t_colptr, *t_rowind;	/* a column oriented form of T = A' */

  marker = (int*) SP_Malloc ((MAX(m,n)+1)*sizeof(int));
  thiscol = (int*) SP_Malloc (m*sizeof(int));
  t_colptr = (int*) SP_Malloc ((m+1)*sizeof(int));
  t_rowind = (int*) SP_Malloc (nz*sizeof(int));
    
  /* Get counts of each column of T, and set up column pointers */
  for (i = 0; i < m; ++i) marker[i] = 0;
  for (j = 0; j < n; ++j) {
    for (i = colptr[j]; i < colptr[j+1]; ++i)
      ++marker[rowind[i]];
  }
  t_colptr[0] = 0;
  for (i = 0; i < m; ++i) {
    t_colptr[i+1] = t_colptr[i] + marker[i];
    marker[i] = t_colptr[i];
  }

  /* Transpose the matrix from A to T */
  for (j = 0; j < n; ++j)
    for (i = colptr[j]; i < colptr[j+1]; ++i) {
      col = rowind[i];
      t_rowind[marker[col]] = j;
      ++marker[col];
    }

  /* Zero the diagonal flag */
  for (i = 0; i < n; ++i) marker[i] = -1;
  
  /* First pass determines number of nonzeros in B */
  num_nz = 0;
  for (j = 0; j < n; ++j) {
    /* Flag the diagonal so it's not included in the B matrix */
    marker[j] = j;
    
    for (i = colptr[j]; i < colptr[j+1]; ++i) {
      /* A_kj is nonzero, add pattern of column T_*k to B_*j */
      k = rowind[i];
      for (ti = t_colptr[k]; ti < t_colptr[k+1]; ++ti) {
	trow = t_rowind[ti];
	if ( marker[trow] != j ) {
	  marker[trow] = j;
	  num_nz++;
	}
      }
    }
  }
  *atanz = num_nz;
    
  /* Allocate storage for A'*A */
  *ata_colptr = (int*) SP_Malloc((n+1)*sizeof(int));

  if ( *atanz ) {
    *ata_rowind = (int*) SP_Malloc((*atanz)*sizeof(int));
  }
  b_colptr = *ata_colptr; /* aliasing */
  b_rowind = *ata_rowind;
    
  /* Zero the diagonal flag */
  for (i = 0; i < n; ++i) marker[i] = -1;
    
  /* Compute each column of B, one at a time */
  num_nz = 0;
  for (j = 0; j < n; ++j) {
    b_colptr[j] = num_nz;
	
    /* Zero this column */
    for (i = 0; i < m; ++i) thiscol[i] = 0;

    /* Flag the diagonal so it's not included in the B matrix */
    marker[j] = j;

    for (i = colptr[j]; i < colptr[j+1]; ++i) {
      /* A_kj is nonzero, add pattern of column T_*k to B_*j */
      k = rowind[i];
      for (ti = t_colptr[k]; ti < t_colptr[k+1]; ++ti) {
	trow = t_rowind[ti];
	if ( marker[trow] != j ) {
	  marker[trow] = j;
	  thiscol[trow] = 1;
	}
      }
    }

    /* Add all elements in this column */
    for (i = 0; i < m; ++i) 
      if (thiscol[i])
	b_rowind[num_nz++] = i;
  }
  b_colptr[n] = num_nz;
       
  SP_Free(marker);
  SP_Free(thiscol);
  SP_Free(t_colptr);
  SP_Free(t_rowind);
}


static void
_construct_at_plus_a (
       const int m,      /* INPUT - number of rows in matrix A. */
       const int n,      /* INPUT - number of columns in matrix A. */
       const int nz,     /* INPUT - number of nonzeros in matrix A */
       int *colptr,      /* INPUT - column pointer of size n+1 for matrix A. */
       int *rowind,      /* INPUT - row indices of size nz for matrix A. */
       int *atanz,       /* OUTPUT - on exit, returns the actual number of
                            nonzeros in matrix A'+A. */
       int **ata_colptr, /* OUTPUT - size n+1 */
       int **ata_rowind  /* OUTPUT - size *atanz */
       )
/*
 * Purpose
 * =======
 *
 * Form the structure of A'+A. A is an m-by-n matrix in column oriented
 * format represented by (colptr, rowind). The output A'+A is in column
 * oriented format (symmetrically, also row oriented), represented by
 * (ata_colptr, ata_rowind).
 */
{
  register int i, j, col, num_nz, ti;
  int *marker, *thiscol, *b_colptr, *b_rowind;
  int *t_colptr, *t_rowind;	/* a column oriented form of T = A' */

  marker = (int*) SP_Malloc ((MAX(m,n)+1)*sizeof(int));
  thiscol = (int*) SP_Malloc (m*sizeof(int));
  t_colptr = (int*) SP_Malloc ((m+1)*sizeof(int));
  t_rowind = (int*) SP_Malloc (nz*sizeof(int));
    
  /* Get counts of each column of T, and set up column pointers */
  for (i = 0; i < m; ++i) marker[i] = 0;
  for (j = 0; j < n; ++j) {
    for (i = colptr[j]; i < colptr[j+1]; ++i)
      ++marker[rowind[i]];
  }
  t_colptr[0] = 0;
  for (i = 0; i < m; ++i) {
    t_colptr[i+1] = t_colptr[i] + marker[i];
    marker[i] = t_colptr[i];
  }

  /* Transpose the matrix from A to T */
  for (j = 0; j < n; ++j)
    for (i = colptr[j]; i < colptr[j+1]; ++i) {
      col = rowind[i];
      t_rowind[marker[col]] = j;
      ++marker[col];
    }

  /* Zero the diagonal flag */
  for (i = 0; i < n; ++i) marker[i] = -1;
  
  /* First pass determines number of nonzeros in B */
  num_nz = 0;
  for (j = 0; j < n; ++j) {
    /* Zero this column */
    for (i = 0; i < m; ++i) thiscol[i] = 0;

    /* add pattern of column A_*j to thiscol */
    for (i = colptr[j]; i < colptr[j+1]; ++i)
      thiscol[rowind[i]] = 1;
    
    /* add pattern of column T_*j to thiscol */
    for (ti = t_colptr[j]; ti < t_colptr[j+1]; ++ti)
      thiscol[t_rowind[ti]] = 1;

    /* the diagonal is not included in the B matrix */
    thiscol[j] = 0;

    for (i = 0; i < m; ++i) 
      if (thiscol[i])
	num_nz++;    
  }
  *atanz = num_nz;
    
  /* Allocate storage for A'+A */
  *ata_colptr = (int*) SP_Malloc((n+1)*sizeof(int));

  if ( *atanz ) {
    *ata_rowind = (int*) SP_Malloc((*atanz)*sizeof(int));
  }
  b_colptr = *ata_colptr; /* aliasing */
  b_rowind = *ata_rowind;
    
  /* Zero the diagonal flag */
  for (i = 0; i < n; ++i) marker[i] = -1;
    
  /* Compute each column of B, one at a time */
  num_nz = 0;
  for (j = 0; j < n; ++j) {
    b_colptr[j] = num_nz;

    /* Zero this column */
    for (i = 0; i < m; ++i) thiscol[i] = 0;

    /* add pattern of column A_*j to thiscol */
    for (i = colptr[j]; i < colptr[j+1]; ++i)
      thiscol[rowind[i]] = 1;
    
    /* add pattern of column T_*j to thiscol */
    for (ti = t_colptr[j]; ti < t_colptr[j+1]; ++ti)
      thiscol[t_rowind[ti]] = 1;

    /* the diagonal is not included in the B matrix */
    thiscol[j] = 0;

    for (i = 0; i < m; ++i) 
      if (thiscol[i])
	b_rowind[num_nz++] = i;    
  }
  b_colptr[n] = num_nz;
       
  SP_Free(marker);
  SP_Free(thiscol);
  SP_Free(t_colptr);
  SP_Free(t_rowind);
}


static void
_No0Diagonal (
	SP_SparseMatrix *A,	/* INPUT - sparse matrix */
	int *perm_c,		/* INPUT - column permutation vector currently associated with A */
	int *perm_r		/* OUTPUT - row permutation vector eliminating zero diagonals */
	) 
/*
 * Purpose
 * =======
 * 
 * Generating row permutation vector to eliminate zero diagonal elements.
 * This routine performs a symbolic factorization.
 */
{
  int i, k, k1, tmpi, tmpk, cursor, curCol, row, col;
  int *Iperm_r;
  NodeList** rows;

  Iperm_r = (int *) SP_Malloc(A->order*sizeof(int));
  for (i=0; i<A->order; i++) perm_r[i] = Iperm_r[i] = i;

  /* form the original symbolic matrix */
  rows = (NodeList**) SP_Malloc(A->order * sizeof(NodeList*));
  for (i=0; i<A->order; i++)
    rows[i] = NULL;

  curCol = 1;
  for (cursor=0; cursor<A->numNZ; cursor++) {
    row = A->rowIndex[cursor];

    if (cursor == A->colStart[curCol]) curCol++;
    col = curCol-1;

    /* permute col */
    col = perm_c[col];

    SP_InsertToColumn(rows+row, col, col);
  }

  for (k=0; k<A->order; k++) {
    int firstrow = -1;
    NodeList* firstrowNL = NULL;
    for (k1=k; k1<A->order; k1++) {
      /* whether row k1 has a nonzero in column k */
      NodeList* tmpNL = rows[k1];
      while (tmpNL!=NULL) {
	NodeList* tmpNLnext;
	if (tmpNL->from>k) { /* over-shoot */
	  tmpNL = NULL;
	  break;
	}
	if (tmpNL->from<=k && tmpNL->to>=k) { /* match */
	  break;
	}

	tmpNLnext = tmpNL->next;
	rows[k1] = tmpNLnext;
	SP_FreeNL(tmpNL);
	tmpNL = tmpNLnext;
      }

      if (tmpNL!=NULL) {  /* row k1 has a nonzero in column k */
	if (firstrow==-1) {
	  firstrow = k1;
	  firstrowNL = tmpNL;
	}
	else {
	  /* copy row firstrow (ROUGHLY from the diagonal) to row k1 */
	  tmpNL = firstrowNL;
	  while (tmpNL!=NULL) {
	    SP_InsertToColumn(rows+k1, tmpNL->from, tmpNL->to);
	    tmpNL = tmpNL->next;
	  }
	}
      }
    }

    if (firstrow == -1)
      SP_err(SP_ERR_LOC, "Could not find row permutation to have nonzero diagonal: matrix is singular");

    i = firstrow;
    if (i!=k) {	/* exchange row i and k */
      NodeList* tmpNL = rows[k];
      rows[k] = rows[i];
      rows[i] = tmpNL;

      tmpi = Iperm_r[i];
      tmpk = Iperm_r[k];
      perm_r[tmpi] = k;
      perm_r[tmpk] = i;
      Iperm_r[i] = tmpk;
      Iperm_r[k] = tmpi;
    }
  }

  /* Free the temporary space */
  SP_Free(Iperm_r);
  SP_Free(rows);
  SP_FreeAllNL();
}


static void
_Min0Diagonal (
	SP_SparseMatrix *A,	/* INPUT - sparse matrix */
	int *perm_c,		/* INPUT - column permutation vector currently associated with A */
	int *perm_r		/* OUTPUT - row permutation vector minizing zero diagonals */
	) 
/*
 * Purpose
 * =======
 * 
 * Generating row permutation vector to minimize the number of zero diagonal elements.
 * The algorithm to eliminate all zero diagonal elements has exponential complexity and
 * we use an approximation algorithm here. Further enhancement is possible.
 */
{
  int i, k, tmpi, tmpk, cursor, temprow, min_num, min_row = -1;
  int *Iperm_c, *Iperm_r, *RowAssigned, *LeftOver;

  Iperm_r = (int *) SP_Malloc(A->order*sizeof(int));
  for (i=0; i<A->order; i++) perm_r[i] = Iperm_r[i] = i;

  Iperm_c = (int *) SP_Malloc(A->order*sizeof(int));
  for (i= 0; i<A->order; i++) Iperm_c[perm_c[i]] = i;

  RowAssigned = (int *) SP_Malloc (A->order*sizeof(int));
  bzero (RowAssigned, A->order*sizeof(int));

  LeftOver= (int *) SP_Malloc (A->order*sizeof(int));
  bzero (LeftOver, A->order*sizeof(int));
  for (i=0; i<A->numNZ; i++) LeftOver[A->rowIndex[i]]++;

  for (k=0; k<A->order; k++) {
    /* find the row with minimus leftover elements */
    min_num=A->order;

    cursor = A->colStart[Iperm_c[k]];
    while (cursor< A->colStart[Iperm_c[k]+1]) {
      temprow = perm_r[A->rowIndex[cursor]];
      if (!RowAssigned[temprow]) {
	LeftOver[A->rowIndex[cursor]]--;
	if (LeftOver[A->rowIndex[cursor]]<min_num) {
	  min_row=temprow;
	  min_num=LeftOver[A->rowIndex[cursor]];
	}
      }
      cursor++;
    }

    if (min_num!=A->order) {	/* find something */
      i=min_row;
      if (i!=k) {	/* exchange row i and k */
	tmpi = Iperm_r[i];
	tmpk = Iperm_r[k];
	perm_r[tmpi] = k;
	perm_r[tmpk] = i;
	Iperm_r[i] = tmpk;
	Iperm_r[k] = tmpi;
      }
      RowAssigned[k] = 1;
    }
    /* we could do some recursive search if the initial probe fails, we leave this as future work */
  }

  /* Free the space uses by the working arrays */
  SP_Free(Iperm_r);
  SP_Free(Iperm_c);
  SP_Free(RowAssigned);
  SP_Free(LeftOver);
}


void 
SP_Ordering (
	int mode,		/* INPUT - specifies the way ordering is conducted:
				 * mode==0: natual ordering (i.e., Pc=I, Pr=I)
				 * mode==1: MMD ordering of the structure of A'*A and 
				 *          a row permutation to minimize zero diagonals
				 */
	SP_SparseMatrix *A,	/* INPUT - sparse matrix to be ordered */
	int **perm_c,		/* OUTPUT - column permutation vector of size A->order, 
				 * which defines the permutation matrix Pc; perm_c[i] = j 
				 * means column i of A is in position j in A*Pc.
				 */
	int **perm_r		/* OUTPUT - row permutation vector of size A->order */
	)
/*
 * Purpose
 * =======
 *
 * SP_Ordering obtains permutation vectors Pc and Pr. The LU factorization of A*Pc tends to 
 * have less fill-in than the LU factorization of A. 
 */
{
  int n=A->order, bnz, *b_colptr, i;
  int delta, maxint, nofsub, *invp;
  int *b_rowind, *dhead, *qsize, *llist, *marker;

  *perm_c = (int*) SP_Malloc (n*sizeof(int));
  *perm_r = (int*) SP_Malloc (n*sizeof(int));

  switch (mode) {	
  case 0:	/* Natural ordering */
    for (i=0; i<n; ++i) (*perm_c)[i] = i;
    break;
  
  case 1:	/* Minimum degree ordering on A'*A */
  case 2:       /* Minimum degree ordering on A'+A */
    if (mode==1)
      _construct_at_times_a (n, n, A->numNZ, A->colStart, A->rowIndex, &bnz, &b_colptr, &b_rowind);
    if (mode==2)
      _construct_at_plus_a (n, n, A->numNZ, A->colStart, A->rowIndex, &bnz, &b_colptr, &b_rowind);

    if ( bnz != 0 ) {
      /* Initialize and allocate storage for GENMMD. */
      delta = 1; 
      maxint = 2147483647; /* 2**31 - 1 */
      invp = (int *) SP_Malloc((n+delta)*sizeof(int));
      dhead = (int *) SP_Malloc((n+delta)*sizeof(int));
      qsize = (int *) SP_Malloc((n+delta)*sizeof(int));
      llist = (int *) SP_Malloc(n*sizeof(int));
      marker = (int *) SP_Malloc(n*sizeof(int));

      /* Transform adjacency list into 1-based indexing required by GENMMD.*/
      for (i=0; i<=n; ++i) ++b_colptr[i];
      for (i=0; i<bnz; ++i) ++b_rowind[i];
	
      genmmd_ (&n, b_colptr, b_rowind, *perm_c, invp, &delta, dhead, qsize, llist, marker, &maxint, &nofsub);
      
      /* Transform perm_c into 0-based indexing. */
      for (i=0; i<n; ++i) --(*perm_c)[i];

      SP_Free (b_colptr);
      SP_Free (b_rowind);
      SP_Free (invp);
      SP_Free (dhead);
      SP_Free (qsize);
      SP_Free (llist);
      SP_Free (marker);
    } else {
      for (i=0; i<n; ++i) (*perm_c)[i] = i;
    }
    break;

  default:
    SP_err(SP_ERR_LOC, "Invalid ordering mode");
  }

  /* The main goal of the row permutation is to minimize diagonal zero elements.
     This is because diagonal zeros cause over-estimation during the subsequent
     symbolic factorization.
  */
  if (SP_SparseMatrixDiagonalZero(A) == 0) {
    /* If the original matrix has no diagonal zeros, then making perm_r==perm_c
       eliminates diagonal zeros.
    */
    for (i=0; i<n; ++i) (*perm_r)[i] = (*perm_c)[i];
  }
  else {
    _No0Diagonal (A, *perm_c, *perm_r);
  }
}
