/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * SnMatrix.c - operations associated with the supernodal matrix
 *
 */


#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <strings.h>

#include "struct.h"
#include "functions.h"
#include "globals.h"

/* Release the memory occupied by supernode matrix - internal matrix representation */
void SP_DestroySNmatrix(SP_SNmatrix_struct *SNma) {
  int i;

  SP_Free (SNma->Snode);
  SP_Free (SNma->Sstart);
  SP_Free (SNma->SnodeChild);

  for (i=0; i<SNma->LstartCol[SNma->colNum]; i++)
    if (SNma->LnodeCol[i]) SP_Free (SNma->LnodeCol[i]);
  SP_Free (SNma->LstartCol);
  SP_Free (SNma->LindexCol);
  SP_Free (SNma->LsubRow);
  SP_Free (SNma->LsubRowNumInBlk);
  SP_Free (SNma->LsubRowNumInCol);

  SP_Free (SNma->LstartRow);
  SP_Free (SNma->LindexRow);
  SP_Free (SNma->Lrow2col);
  SP_Free (SNma->LcacheRow);
  
  for (i=0; i<SNma->UstartRow[SNma->rowNum]; i++)
    if (SNma->UnodeRow[i]) SP_Free (SNma->UnodeRow[i]);
  SP_Free (SNma->UstartRow);
  SP_Free (SNma->UindexRow);
  SP_Free (SNma->UsubCol);
  SP_Free (SNma->UsubColNumInBlk);
  SP_Free (SNma->UsubColNumInRow);

  SP_Free (SNma->UstartCol);
  SP_Free (SNma->UindexCol);
  SP_Free (SNma->Ucol2row);

  SP_Free (SNma);
}

extern int SP_cbufsize, SP_rbufsize;
extern int SP_ibufsize;

INDEXTYPE SP_maxNumColSendBlks=0;
INDEXTYPE SP_maxNumRowSendBlks=0;

/* find the columnwise index of block (blkRow, blkCol) in L part */
static INDEXTYPE _LColfindSnode(SP_SNmatrix_struct *SNma, INDEXTYPE blkRow, INDEXTYPE blkCol) {
  INDEXTYPE low, high, mid;

  low = SNma->LstartCol[CINDEX(blkCol)];
  high = SNma->LstartCol[CINDEX(blkCol)+1]-1;

  while (low <= high) {
    mid = (low+high)/2;
    if (SNma->LindexCol[mid] == blkRow) return mid;
    if (SNma->LindexCol[mid] < blkRow) low = mid+1;
    else high = mid-1;
  }
  return -1;
} 

/* find the rowwiseindex of block (blkRow, blkCol) in U part */
static INDEXTYPE _URowfindSnode(SP_SNmatrix_struct *SNma, INDEXTYPE blkRow, INDEXTYPE blkCol) {
  INDEXTYPE low, high, mid;

  low = SNma->UstartRow[RINDEX(blkRow)];
  high = SNma->UstartRow[RINDEX(blkRow)+1]-1;

  while (low <= high) {
    mid = (low+high)/2;
    if (SNma->UindexRow[mid] == blkCol) return mid;
    if (SNma->UindexRow[mid] < blkCol) low = mid+1;
    else high = mid-1;
  }
  return -1;
}

/* find the address of element (row, col) */
static ELEMTYPE *_FindElem(SP_SNmatrix_struct *SNma, INDEXTYPE row, INDEXTYPE col, int alloc) {
  INDEXTYPE blkRow, blkCol;
  static INDEXTYPE lastBlkRow = 0, lastBlkCol = 0;
  static INDEXTYPE blkIndex;
  ELEMTYPE *elemAddr;

  blkRow = index2Snode(SNma, row);
  blkCol = index2Snode(SNma, col);

  if (blkRow >= blkCol) {
    if ((blkRow != lastBlkRow) || (blkCol != lastBlkCol)) {
      blkIndex = _LColfindSnode(SNma, blkRow, blkCol);
      lastBlkRow = blkRow;
      lastBlkCol = blkCol;
    }

    if (blkIndex==-1) {
      if (alloc)
	SP_err(SP_ERR_LOC, "Error: row %d, col %d is invalid element!\n", row, col);
      else
	return NULL;
    }
    if (SNma->LnodeCol[blkIndex]==NULL) {
      if (alloc)
	SNma->LnodeCol[blkIndex] = SP_MallocElem( SNma->LsubRowNumInBlk[blkIndex] * DSIZE(SNma, blkCol) );
      else
	return NULL;
    }
    elemAddr = SNma->LnodeCol[blkIndex]; 
    elemAddr += (col-SNma->Sstart[blkCol]) * SNma->LsubRowNumInBlk[blkIndex];
    if (!testcol(SNma->LsubRow[blkIndex], row-SNma->Sstart[blkRow])) {
      if (alloc)
	SP_err(SP_ERR_LOC, "Error: row %d, col %d is invalid element!\n", row, col);
      else
	return NULL;
    }
    elemAddr += numAhead(row-SNma->Sstart[blkRow], SNma->LsubRow[blkIndex]);

    return elemAddr;
  }
  else {
    if ((blkRow != lastBlkRow) || (blkCol != lastBlkCol)) {
      blkIndex = _URowfindSnode(SNma, blkRow, blkCol);
      lastBlkRow = blkRow;
      lastBlkCol = blkCol;
    }

    if (blkIndex==-1) {
      if (alloc)
	SP_err(SP_ERR_LOC, "Error: row %d, col %d is invalid element!\n", row, col);
      else
	return NULL;
    }
    if (SNma->UnodeRow[blkIndex]==NULL) {
      if (alloc)
	SNma->UnodeRow[blkIndex] = SP_MallocElem( SNma->UsubColNumInBlk[blkIndex] * DSIZE(SNma, blkRow) );
      else
	return NULL;
    }
    elemAddr = SNma->UnodeRow[blkIndex]; 
    if (!testcol(SNma->UsubCol[blkIndex], col-SNma->Sstart[blkCol])) {
      if (alloc)
	SP_err(SP_ERR_LOC, "Error: row %d, col %d is invalid element!\n", row, col);
      else
	return NULL;
    }
    elemAddr += numAhead(col-SNma->Sstart[blkCol], SNma->UsubCol[blkIndex]) * DSIZE(SNma, blkRow);
    elemAddr += row - SNma->Sstart[blkRow];

    return elemAddr;
  }
}

/* return the value of entry (row, col), no matter it is allocated or not */
static ELEMTYPE _EntryValue(SP_SNmatrix_struct *SNma, int row, int col, int peroot) {
  MPI_Status SP_status;
  INDEXTYPE blkRow = index2Snode(SNma, row);
  INDEXTYPE blkCol = index2Snode(SNma, col);
  int pe = PEOWNER(blkRow, blkCol);
  ELEMTYPE *addr;
  ELEMTYPE value;

  if (SP_me_no == peroot) {
    if (SP_me_no == pe) {
      addr = _FindElem(SNma, row, col, 0);
      if (addr) return *addr;
      else return 0.0;
    }
    else {
      MPI_Recv(&value, 1, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
      return value;
    }
  }
  else 
    if (SP_me_no == pe ) {
      addr = _FindElem(SNma, row, col, 0);
      if (addr) value=*addr; else value=0.0;
      MPI_Send(&value, 1, ELEMTYPE_MPI, peroot, SP_MPI_TAG, MPI_COMM_WORLD);
    }
  return 0;
}

/* for startrow, endrow, startcol, endcol:  0 means no limit */
void SP_PrintInternalMatrix(SP_SNmatrix_struct *SNma, int peroot, 
			    int startrow, int endrow, int startcol, int endcol) 
{
  int i, j;
  ELEMTYPE val;

  if (startrow==0) startrow = 1;
  if (endrow==0) endrow = SNma->eleOrder;
  if (startcol==0) startcol = 1;
  if (endcol==0) endcol = SNma->eleOrder;

  for (i=startrow; i<=endrow; i++) {
    if (SP_me_no==peroot) fprintf(stdout, "Row %d:", i);
    for (j=startcol; j<=endcol; j++) {
      val = _EntryValue(SNma, i, j, peroot);
      if (SP_me_no==peroot) fprintf(stdout, " %3.3f(%d)", val, PEOWNER(index2Snode(SNma,i),index2Snode(SNma,j)));
    }
    if (SP_me_no==peroot) fprintf(stdout, "\n");
  }

  if (SP_me_no==peroot) fflush(stdout);
}

/* allocate the space for the matrix */
static void _AllocSpace(SP_SNmatrix_struct *SNma) {
  INDEXTYPE i, width, blk;
  INDEXTYPE j;
  int isize, bsize;

  SNma->LnodeCol = (ELEMTYPE**) SP_Malloc(SNma->LstartCol[SNma->colNum] * sizeof(ELEMTYPE*));  
  for (i=0; i<SNma->colNum; i++) {
    width = DSIZE(SNma, i*SP_cp+SP_me_cno+1);
    if ((bsize=width*SNma->LsubRowNumInCol[i]) > SP_cbufsize) SP_cbufsize = bsize;
    if ((bsize=SNma->LstartCol[i+1]-SNma->LstartCol[i]) > SP_maxNumColSendBlks) SP_maxNumColSendBlks = bsize;    
    for (j=SNma->LstartCol[i]; j<SNma->LstartCol[i+1]; j++)
      SNma->LnodeCol[j] = NULL;
  }
  
  /* allocate space for diagonal blocks */
  for (i=1; i<=SNma->snOrder; i++) 
    if (PEOWNER(i,i)==SP_me_no) {
      blk = SNma->LstartCol[CINDEX(i)];
      SNma->LnodeCol[blk] = SP_MallocElem(DSIZE(SNma, i)*DSIZE(SNma, i));
    }

  SNma->UnodeRow = (ELEMTYPE**) SP_Malloc(SNma->UstartRow[SNma->rowNum] * sizeof(ELEMTYPE*));
  for (i=0; i<SNma->rowNum; i++) {
    width = DSIZE(SNma, i*SP_rp+SP_me_rno+1);
    if ((bsize=width*SNma->UsubColNumInRow[i]) > SP_rbufsize) SP_rbufsize = bsize;
    if ((bsize=SNma->UstartRow[i+1]-SNma->UstartRow[i]) > SP_maxNumRowSendBlks) SP_maxNumRowSendBlks = bsize;
    for (j=SNma->UstartRow[i]; j<SNma->UstartRow[i+1]; j++)
      SNma->UnodeRow[j] = NULL;
  }

  for (i=SP_me_rno+1; i<=SNma->snOrder; i+=SP_rp) {
    isize = 1;
    for (j=SNma->LstartRow[RINDEX(i)]; j<SNma->LstartRow[RINDEX(i)+1]; j++)
      isize += DSIZE(SNma, SNma->LindexRow[j])+2;
    for (j=SNma->UstartRow[RINDEX(i)]; j<SNma->UstartRow[RINDEX(i)+1]; j++)
      isize += DSIZE(SNma, SNma->UindexRow[j])+2;
    if (isize > SP_ibufsize) SP_ibufsize=isize;
  }
}

#define BatchSize 128

/* allocate and initialize buffers and data in the matrix */
void SP_InitData (SP_SNmatrix_struct *SNma, SP_SparseMatrix *input, int *perm_c, int *perm_r) {
  MPI_Status SP_status;
  int pe, i, more;
  INDEXTYPE blkRow, blkCol, curCol, row, col;
  ELEMTYPE val;
  struct ENTRY {
    INDEXTYPE row, col;
    ELEMTYPE val;
  } *entry;
  int *index;

  _AllocSpace(SNma);

  if (SP_me_no == 0) {
    entry = (struct ENTRY*) SP_Malloc (BatchSize*SP_nproc*sizeof(struct ENTRY));
    index = (int*) SP_Malloc (SP_nproc*sizeof(int));
    bzero(index, SP_nproc*sizeof(int));

    /* initialize data for all others */
    curCol = 1;
    for (i=0; i<input->numNZ; i++) {
      val = input->coe[i];
      row = input->rowIndex[i]+1;

      if (i == input->colStart[curCol]) curCol++;
      col = curCol;

      /* permute row and col */
      col = perm_c[col-1]+1;
      row = perm_r[row-1]+1;

      blkRow = index2Snode(SNma, row);
      blkCol = index2Snode(SNma, col);

      if ((pe = PEOWNER(blkRow, blkCol)) != 0) {
	more = pe*BatchSize+index[pe];
	entry[more].row = row;
	entry[more].col = col;
	entry[more].val = val;
	
	if (++index[pe] == BatchSize) {
	  MPI_Send( entry+pe*BatchSize, BatchSize*sizeof(struct ENTRY), MPI_CHAR, pe, SP_MPI_TAG, MPI_COMM_WORLD );
	  index[pe] = 0;
	}
      }
      else
	*_FindElem(SNma, row, col, 1) = val;
    }

    for (pe=1; pe<SP_nproc; pe++) {
      entry[pe*BatchSize+index[pe]].row = 0;
      MPI_Send( entry+pe*BatchSize, (index[pe]+1)*sizeof(struct ENTRY), MPI_CHAR, pe, SP_MPI_TAG, MPI_COMM_WORLD );
    }

    SP_Free(entry);
    SP_Free(index);
  }
  else {
    entry = (struct ENTRY*) SP_Malloc (BatchSize*sizeof(struct ENTRY));
    more = 1;
    while (more) {
      MPI_Recv( entry, BatchSize*sizeof(struct ENTRY), MPI_CHAR, 0, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status );
      for (i=0; i<BatchSize; i++) 
	if (entry[i].row==0) {
	  more=0;
	  break;
	}
	else
	  *_FindElem(SNma, entry[i].row, entry[i].col, 1) = entry[i].val;
    }
    SP_Free(entry);
  }
}
