/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * Solve.c - Triangular solve
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "struct.h"
#include "functions.h"
#include "globals.h"

/* RhsBuf stores the B vector. RhsBufSize means the total number of elments in B stored in this processor*/
static ELEMTYPE  *RhsBuf;
static INDEXTYPE RhsBufSize;

/* VecBuf is used to store b[i]'s, later send them to the processors on the same col */
static ELEMTYPE  *VecBuf;

/* blocksize denotes the height of b[i] blocks on this processors
   blockindex denotes the number of such blocks and blockseq[i] means
   the position of block i  */
static INDEXTYPE *blocksize;
static INDEXTYPE *blockseq;
static INDEXTYPE blockindex;

static int _RHScolNum;
static ELEMTYPE *blasbuf;

/* send/recv buffers, used in _InterchangeRHS */
static ELEMTYPE *_sendBuf, *_recvBuf;

#define ELEMENT(m, row, col) (m->element[(col)*m->rowNum+(row)])

/*
static void _PrintRHS() {
  int i;
  fprintf(stdout, "Proc=%d Height=%d RHS: ", SP_me_no, RhsBufSize/_RHScolNum);
  for (i=0; i<RhsBufSize; i++) 
    fprintf(stdout, "%3.3f ", RhsBuf[i]);
  fprintf(stdout, "\n");
  fflush(stdout);
}
*/

static void _Distrib(SP_SNmatrix_struct *SNma, SP_DenseMatrix* B) {
  MPI_Status SP_status;
  int i, j, height, procnum; 
  ELEMTYPE *startB, *startRhs;

  /* allocate space for everybody */  
  blockindex=0;
  blocksize  = (INDEXTYPE*) SP_Malloc((SNma->rowNum)*sizeof(INDEXTYPE));
  blockseq  = (INDEXTYPE*) SP_Malloc((SNma->rowNum+1)*sizeof(INDEXTYPE));
  blockseq[0] = 0;
  for(i=1;i<=SNma->snOrder;i++)
    if(ROWNER(i)==SP_me_rno) {
      blocksize[blockindex] = SNma->Sstart[i+1] - SNma->Sstart[i];
      blockseq[blockindex+1]  = blockseq[blockindex]+ blocksize[blockindex]*_RHScolNum;
      blockindex++;
    }

  RhsBufSize = blockseq[blockindex];
  RhsBuf = (ELEMTYPE*) SP_Malloc(RhsBufSize*sizeof(ELEMTYPE));
  startRhs = RhsBuf;
  
  /* distribute the right hand side matrix to the first column processors */
  if (SP_me_cno==0) {	/* only  processors of the first column involves B at first */
    if (SP_me_no==0) {	/* 0 processor will send B out */
      startB = B->element;
      for (i=1; i<=SNma->snOrder; i++) {
	procnum = ROWNER(i);
	height = SNma->Sstart[i+1]-SNma->Sstart[i];
	if (procnum != 0) {
	  for (j=0; j<_RHScolNum; j++) 
	    memcpy(VecBuf+j*height, startB+j*B->rowNum, height*sizeof(ELEMTYPE));
	  MPI_Send(VecBuf, height*_RHScolNum, ELEMTYPE_MPI, procnum, SP_MPI_TAG, MPI_COMM_WORLD);
	}
	else {		/* keep those parts of B in processor 0 */
	  for (j=0; j<_RHScolNum; j++) 
	    memcpy(startRhs+j*height, startB+j*B->rowNum, height*sizeof(ELEMTYPE));
	  startRhs += height*_RHScolNum;
	}
	startB += height;
      }
    }
    else { /* other processors in column 1 will recieve B */
      for (i=0; i<blockindex; i++) {
	MPI_Recv(startRhs, blocksize[i]*_RHScolNum, ELEMTYPE_MPI, 0, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
	startRhs += blocksize[i]*_RHScolNum;
      }
    } 
  }
}

/* UTransblock transform a U block into a continguous region stored in buf */
static void _UTransblock(ELEMTYPE *unode, INDEXTYPE height, INDEXTYPE width, PATTYPE pattern, ELEMTYPE *buf){
  register PATTYPE pat=pattern;
  register ELEMTYPE *staunode=unode, *stabuf=buf;

  bzero(buf, width*height*sizeof(ELEMTYPE));

  while (pat) {
    if (pat & 1) {
      memcpy(stabuf, staunode, height*sizeof(ELEMTYPE));
      staunode += height;
    }
    stabuf+=height;
    pat >>= 1;
  }
}

/* only called by processors which own column block blk1 */
static void _InterchangeRHS(SP_SNmatrix_struct *SNma, rowID *Prow, INDEXTYPE blk1) {
  MPI_Status SP_status;
  INDEXTYPE row1, blk2, row2, index1, index2, step1, step2;
  int i, rpno1, rpno2, pe;
  ELEMTYPE tmp;

  rpno1 = ROWNER(blk1);
  for (row1=0; row1<SNma->Sstart[blk1+1]-SNma->Sstart[blk1]; row1++) {
    blk2 = Prow[SNma->Sstart[blk1]+row1].blockNo;
    row2 = Prow[SNma->Sstart[blk1]+row1].rowNo;
    if (blk2==blk1 && row1==row2) continue;
    rpno2 = ROWNER(blk2);

    if (rpno1==rpno2) {		/* local interchange */
      if (rpno1==SP_me_rno) {
	index1 = blockseq[RINDEX(blk1)]+row1;
	index2 = blockseq[RINDEX(blk2)]+row2;
	step1 = SNma->Sstart[blk1+1]-SNma->Sstart[blk1];
	step2 = SNma->Sstart[blk2+1]-SNma->Sstart[blk2];
	for (i=0; i<_RHScolNum; i++) {
	  tmp = RhsBuf[index1];
	  RhsBuf[index1] = RhsBuf[index2];
	  RhsBuf[index2] = tmp;
	  index1+=step1;
	  index2+=step2;
	}
      } 
    }
    else if (rpno1==SP_me_rno) {
      index1 = blockseq[RINDEX(blk1)]+row1;
      step1 = SNma->Sstart[blk1+1]-SNma->Sstart[blk1];
      for (i=0; i<_RHScolNum; i++) _sendBuf[i]=RhsBuf[index1+i*step1];
      pe = SP_me_cno*SP_rp + rpno2;
      MPI_Sendrecv(_sendBuf, _RHScolNum, ELEMTYPE_MPI, pe, SP_MPI_TAG, 
		   _recvBuf, _RHScolNum, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
      for (i=0; i<_RHScolNum; i++) RhsBuf[index1+i*step1]=_recvBuf[i];
    }
    else if (rpno2==SP_me_rno) {
      index2 = blockseq[RINDEX(blk2)]+row2;
      step2 = SNma->Sstart[blk2+1]-SNma->Sstart[blk2];
      for (i=0; i<_RHScolNum; i++) _sendBuf[i]=RhsBuf[index2+i*step2];
      pe = SP_me_cno*SP_rp + rpno1;
      MPI_Sendrecv(_sendBuf, _RHScolNum, ELEMTYPE_MPI, pe, SP_MPI_TAG, 
		   _recvBuf, _RHScolNum, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
      for (i=0; i<_RHScolNum; i++) RhsBuf[index2+i*step2]=_recvBuf[i];
    }
  }
}

static void _LSolve(SP_SNmatrix_struct *SNma, rowID *Prow, SP_DenseMatrix* B) {
  MPI_Status SP_status;
  int i,j, procnum, height, width, trueHeight;
  ELEMTYPE *position;
  ELEMTYPE  *startBlas;
  INDEXTYPE stc, endc;
  PATTYPE pattern;
  ELEMTYPE ONE = 1.0, ZERO = 0.0;
  
  _sendBuf = SP_Malloc(_RHScolNum*sizeof(ELEMTYPE));
  _recvBuf = SP_Malloc(_RHScolNum*sizeof(ELEMTYPE));

  for (i=1;i<=SNma->snOrder;i++) {	/* go through all the columns to find Linverse(B) */
    if (SP_me_cno==COWNER(i)) {		/* the processors of column i  will use b[i] to update b[j] */  
      _InterchangeRHS(SNma, Prow, i);
  
      if(SP_me_rno==ROWNER(i)){/* the processor solving b[i] */
	position = RhsBuf + blockseq[RINDEX(i)];
	height = blocksize[RINDEX(i)];
	spTRSM (SP_ftcL, SP_ftcL, SP_ftcN, SP_ftcU, &height, &_RHScolNum, 
                &ONE, SNma->LnodeCol[SNma->LstartCol[CINDEX(i)]],&height,position,&height); 
        memcpy(VecBuf, position, height*_RHScolNum*sizeof(ELEMTYPE));

	/* send VecBuf to processors in the same col */
	for(j=0, procnum = SP_rp*COWNER(i); j<SP_rp; j++, procnum++)
	  if(procnum != SP_me_no) {
	    MPI_Send(position, height*_RHScolNum, ELEMTYPE_MPI, procnum, SP_MPI_TAG, MPI_COMM_WORLD);
	  }
      }
      else {	/* the processors in the same column receive the revised B[i] */
	MPI_Recv(VecBuf, SP_BlkSize*_RHScolNum, ELEMTYPE_MPI, PEOWNER(i,i),
		 SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
      }

      /* update the blocks in the same column using the changed blocks */
      stc = SNma->LstartCol[CINDEX(i)];
      endc = SNma->LstartCol[CINDEX(i)+1];
  
      while(SNma->LindexCol[stc]<=i && stc<endc) stc++;
      while(stc < endc){
        if(SNma->LnodeCol[stc]!=NULL) {
	  position = RhsBuf + blockseq[RINDEX(SNma->LindexCol[stc])];
	  width  = SNma->Sstart[i+1] - SNma->Sstart[i];
	  height = SNma->LsubRowNumInBlk[stc];
	  pattern = SNma->LsubRow[stc];
	  bzero(blasbuf, height*width*sizeof(ELEMTYPE));
	  spGEMM (SP_ftcN, SP_ftcN, &height, &_RHScolNum, &width, &ONE, SNma->LnodeCol[stc], &height, VecBuf, &width, &ZERO, blasbuf, &height);
	  startBlas = blasbuf;
	  trueHeight = SNma->Sstart[SNma->LindexCol[stc]+1]-SNma->Sstart[SNma->LindexCol[stc]];
	  while (pattern) {
	    if (pattern & 1) {
	      for(j=0;j<_RHScolNum;j++) position[j*trueHeight]-=startBlas[j*height];
	      startBlas++;
	    }
	    pattern >>= 1;
	    position++;
	  }
	}
        stc++;
      }

      /* send the updated RhsBuf to next col */
      if (SP_cp>1 && i<SNma->snOrder) {
	procnum = (SP_me_no + SP_rp) % SP_nproc;
	height = blockseq[RINDEX((i-SP_cp+2)>1?(i-SP_cp+2):1)];
	MPI_Send(RhsBuf+height, RhsBufSize-height, ELEMTYPE_MPI, procnum, SP_MPI_TAG, MPI_COMM_WORLD);
      }
    }
    if (SP_cp>1 && i<SNma->snOrder && SP_me_cno==COWNER(i+1)) { /* receive the updated RhsBuf from previous col */
       procnum = (SP_me_no - SP_rp + SP_nproc) % SP_nproc;
       height = blockseq[RINDEX((i-SP_cp+2)>1?(i-SP_cp+2):1)];
       MPI_Recv(RhsBuf+height, RhsBufSize-height, ELEMTYPE_MPI, procnum, SP_MPI_TAG, 
                MPI_COMM_WORLD, &SP_status);
    }
  }

  SP_Free(_sendBuf);
  SP_Free(_recvBuf);
}

static void _USolve(SP_SNmatrix_struct *SNma, SP_DenseMatrix* B) {
  MPI_Status SP_status;
  int i,j, procnum, height, width;
  ELEMTYPE  *position;
  INDEXTYPE  str, endr;
  ELEMTYPE minusONE = -1.0, ONE = 1.0;
  
  for(i=SNma->snOrder;i>=1;i--){
    if(SP_me_cno==COWNER(i)) {
      if(SP_me_rno==ROWNER(i)) {
	position = RhsBuf + blockseq[RINDEX(i)];
	/* find Uinverse(B_i) for each column of B */
	height = blocksize[RINDEX(i)];
	spTRSM (SP_ftcL, SP_ftcU, SP_ftcN, SP_ftcN, &height, &_RHScolNum, &ONE, 
		SNma->LnodeCol[SNma->LstartCol[CINDEX(i)]],&height,position,&height);
	memcpy(VecBuf, position, height*_RHScolNum*sizeof(ELEMTYPE));
	
	/* send VecBuf to processors in the same col */
	for(j=0, procnum = SP_rp*COWNER(i); j<SP_rp; j++, procnum++)
	  if(procnum != SP_me_no) {
	    MPI_Send(position, height*_RHScolNum, ELEMTYPE_MPI, procnum, 
		     SP_MPI_TAG, MPI_COMM_WORLD);
	  }
      }
      else{ /* other processor in the same col will receive B_i */ 
	MPI_Recv(VecBuf, SP_BlkSize*_RHScolNum, ELEMTYPE_MPI, PEOWNER(i,i),
                 SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
      }

      /* update the blocks in the same column using the changed blocks */
      str = SNma->UstartCol[CINDEX(i)];
      endr = SNma->UstartCol[CINDEX(i)+1]-1;
 
      while(endr >= str){
         if(SNma->UnodeRow[SNma->Ucol2row[endr]]!=NULL) {
	   position = RhsBuf + blockseq[RINDEX(SNma->UindexCol[endr])];
	   width  = SNma->Sstart[i+1] - SNma->Sstart[i];
	   height = blocksize[RINDEX(SNma->UindexCol[endr])];
	   _UTransblock(SNma->UnodeRow[SNma->Ucol2row[endr]], height, width, SNma->UsubCol[SNma->Ucol2row[endr]], blasbuf);
	   spGEMM (SP_ftcN, SP_ftcN, &height, &_RHScolNum, &width, &minusONE, blasbuf, 
                   &height, VecBuf, &width, &ONE, position, &height);
	 }
         endr--;
      }

      if (SP_cp>1 && i>1) {
	height = (i+SP_cp-2)<(SNma->snOrder)?(i+SP_cp-2):(SNma->snOrder);
	while (height>0 && ROWNER(height)!=SP_me_rno) height--;
	if (height>0) {
	  procnum = (SP_me_no - SP_rp + SP_nproc) % SP_nproc;
	  height = blockseq[RINDEX(height)+1];
	  MPI_Send(RhsBuf, height, ELEMTYPE_MPI, procnum, SP_MPI_TAG, MPI_COMM_WORLD);
	}
      }
    }
    if (SP_cp>1 && i>1 && SP_me_cno==COWNER(i-1)) { /* receive the updated RhsBuf from previous col */
      height = (i+SP_cp-2)<(SNma->snOrder)?(i+SP_cp-2):(SNma->snOrder);
      while (height>0 && ROWNER(height)!=SP_me_rno) height--;
      if (height>0) {
	procnum = (SP_me_no + SP_rp) % SP_nproc;
	height = blockseq[RINDEX(height)+1];
	MPI_Recv(RhsBuf, height, ELEMTYPE_MPI, procnum, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
      }
    }
  }
}

static void _Collect(SP_SNmatrix_struct *SNma, SP_DenseMatrix* B) {
  MPI_Status SP_status;
  int i, j, k;
  INDEXTYPE maxBlkNum, maxBufSize, *tmpSizes;
  INDEXTYPE *otherblocksize;
  ELEMTYPE  *tmpRhs, *stmpRhs;
  ELEMTYPE  *startB, *sRhsBuf;

  /* collect the solution from the first column processors into the right hand side matrix B */
  if(SP_me_cno==0) {
    tmpSizes = (INDEXTYPE*) SP_Malloc(SP_rp*2*sizeof(INDEXTYPE));
    if(SP_me_no==0) {
      sRhsBuf = RhsBuf;
      for (i=0; i<blockindex; i++) {
         startB = B->element + SNma->Sstart[i*SP_rp+1]-1;
	 for (k=0; k<_RHScolNum; k++) {
	   memcpy(startB+k*B->rowNum, sRhsBuf, blocksize[i]*sizeof(ELEMTYPE));
	   sRhsBuf += blocksize[i];
	 }
      }
      
      maxBufSize = 0;
      maxBlkNum = 0;
      for (i=1; i<SP_rp; i++) {
	MPI_Recv(tmpSizes+2*i, 2, INDEXTYPE_MPI, i, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
	if (tmpSizes[2*i]>maxBufSize) maxBufSize = tmpSizes[2*i];
	if (tmpSizes[2*i+1]>maxBlkNum) maxBlkNum = tmpSizes[2*i+1];
      }
      tmpRhs = SP_Malloc(maxBufSize*sizeof(ELEMTYPE));
      otherblocksize = SP_Malloc(maxBlkNum*sizeof(INDEXTYPE));
      for (i=1; i<SP_rp; i++) {
	MPI_Recv(tmpRhs, tmpSizes[2*i], ELEMTYPE_MPI, i, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
	MPI_Recv(otherblocksize, tmpSizes[2*i+1], INDEXTYPE_MPI, i, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
	stmpRhs = tmpRhs;
        for (j=0; j<tmpSizes[2*i+1]; j++) {
	  startB = B->element + SNma->Sstart[j*SP_rp+i+1]-1;
	  for (k=0; k<_RHScolNum; k++) {
	    memcpy(startB+k*B->rowNum, stmpRhs, otherblocksize[j]*sizeof(ELEMTYPE));
	    stmpRhs += otherblocksize[j];
	  }
        }
      }
      SP_Free(otherblocksize);
      SP_Free(tmpRhs);
    }
    else{
      tmpSizes[0] = RhsBufSize;
      tmpSizes[1] = blockindex;
      MPI_Send(tmpSizes, 2, INDEXTYPE_MPI, 0, SP_MPI_TAG, MPI_COMM_WORLD);
      MPI_Send(RhsBuf, RhsBufSize, ELEMTYPE_MPI, 0, SP_MPI_TAG, MPI_COMM_WORLD);
      MPI_Send(blocksize, blockindex, INDEXTYPE_MPI, 0, SP_MPI_TAG, MPI_COMM_WORLD);
    } 
    SP_Free(tmpSizes);
  }

  SP_Free(blocksize);
  SP_Free(blockseq);
}

static void _PermuteBack (SP_DenseMatrix *B, int *perm) {
  int i, j, current;
  ELEMTYPE *ex_buf;
  char *marker;

  ex_buf = (ELEMTYPE*) SP_Malloc (B->colNum*sizeof(ELEMTYPE));
  marker = (char*) SP_Malloc (B->rowNum*sizeof(char));
  bzero (marker, B->rowNum*sizeof(char));

  for (i=0; i<B->rowNum; i++) 
    if (marker[i]==0 && perm[i]!=i) {
      for (j=0; j<B->colNum; j++)
	ex_buf[j] = B->element[i+j*B->rowNum];
      current=i;
      while (perm[current]!=i) {
	for (j=0; j<B->colNum; j++) 
	  B->element[current+j*B->rowNum] = B->element[perm[current]+j*B->rowNum];
	current = perm[current];
	marker[current]=1;
      }
      for (j=0; j<B->colNum; j++) 
	B->element[current+j*B->rowNum] = ex_buf[j];
    }

  SP_Free(ex_buf);
  SP_Free(marker);
}

static void _Permute (SP_DenseMatrix *B, int *perm) {
  int i, *new_perm;
  new_perm = (int*) SP_Malloc (B->rowNum*sizeof(int));
  for (i=0; i<B->rowNum; i++) new_perm[perm[i]]=i;
  _PermuteBack (B, new_perm);
  SP_Free (new_perm);
}


void 
SP_TR_Solve (
	SP_SNmatrix_struct *SNma,	/* INPUT - sparse matrix after LU factorization */
	rowID *Prow,			/* INPUT - row interchange vector conducted in LU factorization */
	SP_DenseMatrix *B,		/* INPUT/OUTPUT -  right-hand side matrix,
					   will be overwritten by the result matrix */
	int *perm_c,			/* INPUT - column permutation vector for mmd ordering */
	int *perm_r			/* INPUT - row permutation vector to maintain zero-free 
					   diagonal after mmd ordering */
	) 
/*
 * Purpose
 * =======
 *
 * Trianlgal solve routine, input matrix SNma is assumed having been through LU factorization. 
 */
{
  if (SP_me_no == 0) _RHScolNum=B->colNum;
  MPI_Bcast(&_RHScolNum, 1, MPI_INT, 0, MPI_COMM_WORLD);

  VecBuf =  (ELEMTYPE*) SP_Malloc(SP_BlkSize*_RHScolNum*sizeof(ELEMTYPE));

  if (SP_me_no==0) _Permute(B, perm_r);

  _Distrib(SNma, B);

  blasbuf = SP_Malloc((SP_BlkSize>_RHScolNum?SP_BlkSize:_RHScolNum)*SP_BlkSize*sizeof(ELEMTYPE));

  _LSolve(SNma, Prow, B);

  _USolve(SNma, B);

  SP_Free(blasbuf);
  SP_Free(VecBuf);

  _Collect(SNma, B);

  if (SP_me_no==0) _PermuteBack(B, perm_c);
}
